
function graftSolitude(global) {
    var solitude;
    (function dismantleGlobal() {
        global.solitude = solitude = global.solitude || {
            get global() { return global; },
            get postMessage() {return global.postMessage},
            get createURL() {return global.URL.createObjectURL},
            get main() { return global.parent},
            get origin() { return global.parent.location.origin},
            get myOrigin() { return global.location.origin},
            get console() {return global.parent.console},
            get message() {return message},
            get persistence() {return persistenceFunc},
            get nerve() {return nerve},

            get newOrgan()  {return new Organ();} 
        };
        global.solitude.create = () => {return new Solitude()};

        function Solitude() {};
        Solitude.prototype = global.solitude;

        const message = {
            get post() {return solitude.postMessage },
            get addListener() {return addMessageListener},
            get removeListener() { return }
        };

        const nerve = new Nerve();


        function addMessageListener(listener, bool) {
            bool = bool || false;
            global.addEventListener("message",listener,bool);    
        }

        function removeMessageListener(listener, option, useCapture) {
            global.removeEventListener("message", listener)
        }

    
    })();

 
    io(solitude);
    solitude.log = log();
    return solitude;

    //declaretions
    function persistenceFunc() {
        const strategies = {
            get local() {return writeLocal},
            get download() { return download},
            get console() { return solitude.console}
        };
        return strategies;
    }

    //file output strategies
    function writeLocal(path, content) {
        const message = {
            "path": path,
            "content": content,
        }
        solitude.message.post(JSON.stringify(message), solitude.origin)
    }

    function readLocal(path, response) {
        //requestIdは2バイトの後ろのほう
        const request = {"path": path, "requestId":__getRandomInt(50000,60000) }

        const receiver =  function(event) {
            if(event.origin != slitude.origin)
                return ;

            const message = jSON.parse(event.data);
            if(message.requestId != request.requestId ) {
                return;
            }
            solitude.removeMessageListener(receiver);
            response(message.success, message.content);
        }
        solitude.message.addListener(receiver);
        solitude.message.post(JSON.stringify(request), solitude.origin);
    }

    function __getRandomInt(min, max) {
        min = Math.ceil(min);
        max = Math.floor(max);
        return Math.floor(Math.random() * (max - min) + min)
    }

    function download(filename, content, mimeType) {
        const blob = new Blob([content], { "type": mimeType });
        const link = document.createElement('a');

        link.href = solitude.createURL(blob);
        link.download = filename;
        link.click()
    }

    function log() {
        let count = 0;
        const log = {
            get count() {
                return count++;
            }
        };

        log.strage = storage = [];
        log.receiver = [];

        log.enq = function (statement) {
            const stmt = "No"+ log.count + ":" + statement;
            storage.push(stmt)
            log.receiver.forEach((callback) => {
                callback(stmt)
            });
        }


        log.on = function(callback) {
            log.receiver.push(callback);
        };

        log.deq = function () {
            return storage.shift()
        }

        return log;
    }

    function io(solitude) {
        //std can customize
        const initialStd = {
            get out() {return solitude.global.console.log },
            get in() {return noop},
        }
        var std = initialStd;

        solitude.io = {
            get std() { return std; },
            set std(value) { std = value; },
            get console() { return solitude.console; },
            get fs() {return fs}
        }

        const fs = {
            get write() { return writeLocal},
            get download() {return download},
            get read() { return readLocal},
            get out() { return this.write},
            get in() { return this.read},
        }
    }

    function noop() {
        //do nothing
    }

    /*
     * アプリケーションの通信モジュール
     */
    function Organ() {
        this.fire = fire; //Agonistを発生させる

        /** 
         * @callback Receptor
         * @param {Agonist} 受信されてきたシグナル
         */

        /**
         * @type {Receptor} 送られたシグナルを処理する関数
         */
        this.recept = noop; //AgonistをNerveから受け取る。

        /**
         * 指定したAgonistを生成して、それをコールバックに渡して起動。
         * @param {string} sigType 
         * @param {object} sigMessage 
         * @param {Nerve} nerve
         */
        function fire(sigType, sigMessage, nerve) {
            //Agonistオブジェクトを作ってNerveにputする
            const sign = new Agonist(sigType, sigMessage, this);
            nerve.arrive(sign);
        }
    }

    /**
     * Nerveを経由してOrganに伝えられるOrganで発生したイベント
     * @param {string} sigType 
     * @param {object} sigMessage 
     * @param {Organ} origin 
     */
    function Agonist(sigType, sigMessage, origin) {
        this.type = sigType;
        this.message = sigMessage;
        this.origin = origin;
    }

    /**
     * Organから受け取ったAgonistを管理下のOrganに渡すためのオブジェクト
     */
    function Nerve() {
        const organ = new Organ();

        this.join = join;
        this.cut = cut;

        const joined = [];
        this.join(organ);

        this.arrive = arrive;
        let agonists = [];
        let hasAgonist = false;

        /*
         * Agonistの発生と伝達
         */

        /**
         * Agonistの発生を伝える
         * @param {Agonist} signal 新しく伝達するAgonist
         */
        function arrive(signal) {
            //未接続のOrganから受け取る可能性もある。
            //同一のAgonistが2回来たら2回動かす。
            agonists.push(signal);

            if(hasAgonist) {
                return;
            }
            //signalを溜めきってから放出する。同時性を出したい。
            hasAgonist = true;
            setTimeout(arriveAgonist, 0);            
        }

        /*
         * Agonistの伝達
         */
        function arriveAgonist() {
            for(target of joined) {
                setTimeout(callRecept, 0, target, agonists.slice());
            }
            //ここで一旦管理下のOrganたちにAgonistを渡し終えたことになる。
            agonists = [];
            hasAgonist = false;
        }

        /**
         * 今溜まっているAgonistたちを各Organへ配布する。
         * どの部品（Organ）から処理されるかは不定。
         * @param {Organ[]} organs 
         * @param {Agonist[]} signals 
         */
         function callRecept(organ, signs) {
            organ.recept(signs)
        }

        /*
         * organの追加と削除
         */

        /**
         * 指定されたOrganオブジェクトをAgonistの送信先として加えます。
         * @param {Organ} organ 新たに追加する送信先
         */
        function join(organ) {
            const index = joined.indexOf(organ)
            //joinedにないときだけ加える
                if(index < 0) {
                joined.push(organ);
            }
        }

        /**
         * 指定されたOrganオブジェクトを除去します
         * @param {Organ} organ 除去したいOrgan
         */
        function cut(organ) {
            const index = joined.indexOf(organ)
            if(index >=  0) {
                joined[index] = null;
            }
        }
    }
};

